#include <windows.h>
#include <algorithm>
#include "MQPlugin.h"
#include "MQ3DLib.h"
#include "voxel2.h"

struct RAYHITINFO{
	float fDist;	//C܂łɐi񂾋
	float fProd;	//ʖ@ƃC̓
};

struct RAYHITINFO_SORTER{
	bool operator()(const RAYHITINFO &a,const RAYHITINFO &b){
		return a.fDist<b.fDist;
	}
};

inline float fmax(float x,float y){return x>y? x:y;}
inline float fmin(float x,float y){return x<y? x:y;}


//LȒ_ȂfalseԂ
bool computeBoundaryBox(MQObject lpObject, MQPoint &vMax,MQPoint &vMin)
{
	size_t numVertices=lpObject->GetVertexCount();

	bool bFirst=true;
	for(size_t i=0;i<numVertices;i++){
		if(0!=lpObject->GetVertexRefCount(i)){
			MQPoint v=lpObject->GetVertex(i);
			if(bFirst){
				vMax=v;
				vMin=v;
				bFirst=false;
			}else{
				vMax.x=fmax(vMax.x,v.x);
				vMax.y=fmax(vMax.y,v.y);
				vMax.z=fmax(vMax.z,v.z);

				vMin.x=fmin(vMin.x,v.x);
				vMin.y=fmin(vMin.y,v.y);
				vMin.z=fmin(vMin.z,v.z);
			}
		}
	}
	return !bFirst;
}

//--------------------------------------------------------------------------------------
// Given a ray origin (orig) and direction (dir), and three vertices of a triangle, this
// function returns TRUE and the interpolated texture coordinates if the ray intersects 
// the triangle
//--------------------------------------------------------------------------------------
bool IntersectTriangle(	MQPoint& orig, MQPoint& dir, MQPoint& v0, MQPoint& v1, MQPoint& v2, float* t, float* u, float* v )
{
	// Find vectors for two edges sharing vert0
	MQPoint edge1 = v1 - v0;
	MQPoint edge2 = v2 - v0;

	// Begin calculating determinant - also used to calculate U parameter
	MQPoint pvec=GetExteriorProduct(dir, edge2 );

	// If determinant is near zero, ray lies in plane of triangle
	float det = GetInnerProduct( edge1, pvec );

	MQPoint tvec;
	if( det > 0.0f ){
		tvec = orig - v0;
	}else{
		tvec = v0 - orig;
		det = -det;
	}

	if( det < 0.0001f ){
		return false;
	}

	// Calculate U parameter and test bounds
	*u = GetInnerProduct( tvec, pvec );
	if( *u < 0.0f || *u > det ){
		return false;
	}

	// Prepare to test V parameter
	MQPoint qvec=GetExteriorProduct( tvec, edge1 );

	// Calculate V parameter and test bounds
	*v = GetInnerProduct( dir, qvec );
	if( *v < 0.0f || *u + *v > det ){
		return false;
	}

	// Calculate t, scale parameters, ray intersects triangle
	*t = GetInnerProduct( edge2, qvec );
	float fInvDet = 1.0f / det;
	*t *= fInvDet;
	*u *= fInvDet;
	*v *= fInvDet;

	return true;
}


//Cƃ|Sn_񋓂
void enumHitPointsAlongRay(std::vector<RAYHITINFO> &hitPoints,MQObject lpObject,MQPoint &vOrig,MQPoint &vDir)
{
	hitPoints.clear();

	int numFaces=lpObject->GetFaceCount();
	for(int fidx=0;fidx<numFaces;fidx++){
		int numPoints=lpObject->GetFacePointCount(fidx);
		if(numPoints>2){
			/*
			 * numPoints==0 :폜ς
			 * numPoints==2 :ӂ̂
			 *   ӂƂ͓蔻Ƃ邱ƂłȂ̂ŁAȂ
			 */

			int indices[4];
			lpObject->GetFacePointArray(fidx,indices);

			MQPoint p0=lpObject->GetVertex(indices[0]);
			MQPoint p1=lpObject->GetVertex(indices[1]);
			MQPoint p2=lpObject->GetVertex(indices[2]);

			float fDist,u,v;
			if(IntersectTriangle(vOrig,vDir,p0,p1,p2,&fDist,&u,&v)){
				RAYHITINFO rhi;
				rhi.fDist=fDist;
				rhi.fProd=GetInnerProduct(GetNormal(p0,p1,p2),vDir);
				hitPoints.push_back(rhi);
			}else{
				fDist=-1.0;
			}

			if(numPoints==4){
				float fDistPrev=fDist;
				p0=lpObject->GetVertex(indices[0]);
				p1=lpObject->GetVertex(indices[2]);
				p2=lpObject->GetVertex(indices[3]);
				if(IntersectTriangle(vOrig,vDir,p0,p1,p2,&fDist,&u,&v)){
					if(fDistPrev<-0.5f || fDistPrev!=fDist){
						RAYHITINFO rhi;
						rhi.fDist=fDist;
						rhi.fProd=GetInnerProduct(GetNormal(p0,p1,p2),vDir);
						hitPoints.push_back(rhi);
					}
				}
			}
		}
	}
	std::sort(hitPoints.begin(),hitPoints.end(),RAYHITINFO_SORTER());
}

bool isActiveVoxel(const std::vector<RAYHITINFO> &hitPoints,float fPos)
{
	//NOTE : 悭͂ÃASY͎኱GɂȂ
	//ł͒Pȕ@gɂ

	int flag=0;	//\tO
	for(unsigned int i=0;i<hitPoints.size();i++){
		if(hitPoints[i].fDist > fPos)break;
		if(hitPoints[i].fProd > 0.0f){
			flag++;
		}else{
			flag--;
		}
	}
	return flag%2!=0;
}


//iq\ʂ݂̂ɂ
/*
 * NOTE:
 * {NZiKASYς΂IɌvZ\B
 *   ->\ʂɑ΂Ă̂݃{NZ𐶐悤ȃASY
 * ł͊̃{NZ̉򂩂\ʂ̊iq݂̂𒊏oASYƂ
 */
void ExtractSurface(std::vector<VOXEL> &voxels,int nx,int ny,int nz)
{
	std::vector<BYTE> flags;
	flags.resize(nx*ny*nz);
	memset(&flags[0],0,sizeof(flags[0])*flags.size());

	for(unsigned int i=0;i<voxels.size();i++){
		int x=voxels[i].x;
		int y=voxels[i].y;
		int z=voxels[i].z;

		flags[z*(ny*nx)+y*nx+x]=1;
	}

	std::vector<VOXEL> retArray;
	retArray.reserve(voxels.size());

	//e{NZɂāA͂܂ĂȂXLbv
	for(unsigned int vi=0;vi<voxels.size();vi++){
		int x=voxels[vi].x;
		int y=voxels[vi].y;
		int z=voxels[vi].z;

		if(0<x && x<nx-1 && 0<y && y<ny-1 && 0<z && z<nz-1){
			for(int i=0;i<27;i++){
				//if(i==13)continue;
				int tx=i%3-1;
				int ty=(i/3)%3-1;
				int tz=i/9-1;
				if(tx*tx + ty*ty + tz*tz !=1) continue;	//΂ߕ̓XLȂ

				tx+=x; ty+=y; tz+=z;
				int idx=tz*(ny*nx)+ty*nx+tx;
				if(!flags[idx]){
					//͂܂ĂȂ
					retArray.push_back(voxels[vi]);
					break;
				}
			}
		}else{
			retArray.push_back(voxels[vi]);
		}
	}
	voxels=retArray;
}




#define SCAN_VOXEL(ax,ay,rayAxis)	\
{\
	for(int ay=0;ay<n##ay;ay++){\
		for(int ax=0;ax<n##ax;ax++){\
			MQPoint vOrg;\
			{\
				int rayAxis=0;\
				vOrg=MQPoint(x+0.5f,y+0.5f,z+0.5f)*fSize + vMin - ray*fSize;\
			}\
\
			/*ʂƃCꏊ𒲂ׂ*/\
			enumHitPointsAlongRay(hitPoints,lpObject,vOrg,ray);\
\
			/*Ȃ疳*/\
			if(hitPoints.empty())continue;\
\
			/*eiqɂāAʂ̕\𔻒肷*/\
			for(int rayAxis=0;rayAxis<n##rayAxis;rayAxis++){\
				float pos=(rayAxis+1.0f)*fSize;	/*uC̔˓_vů݂iq̒Sv܂ł̋*/\
				if(isActiveVoxel(hitPoints,pos))voxels.push_back(VOXEL(x,y,z));\
			}\
		}\
	}\
}

bool divideVoxel(std::vector<VOXEL> &voxels,float fSize,MQObject lpObject,int axis,MQPoint &_vMin,BOOL bSurfaceOnly)
{
	MQPoint vMax,vMin;

	//E{bNXZo
	if(!computeBoundaryBox(lpObject,vMax,vMin))return false;

	_vMin=vMin;

	//Zo
	const int nx=int((vMax.x-vMin.x)/fSize)+1;
	const int ny=int((vMax.y-vMin.y)/fSize)+1;
	const int nz=int((vMax.z-vMin.z)/fSize)+1;

	voxels.reserve(nx*ny*nz);

	//CZo
	if(axis==-1){
		axis=0;
		if(nx<ny)axis=1;
		if(ny<nz)axis=2;
	}

	//XL
	std::vector<RAYHITINFO> hitPoints;
	MQPoint ray;
	switch(axis){
	case 0:	//X
		ray=MQPoint(1.0f, 0.0f, 0.0f);
		SCAN_VOXEL(y,z,x);
		break;
	case 1:	//Y
		ray=MQPoint(0.0f, 1.0f, 0.0f);
		SCAN_VOXEL(z,x,y);
		break;
	case 2:	//Z
		ray=MQPoint(0.0f, 0.0f, 1.0f);
		SCAN_VOXEL(x,y,z);
		break;
	default:
		return false;
	}

	if(bSurfaceOnly){
		ExtractSurface(voxels,nx,ny,nz);
	}

	return true;
}




//{NZ𐶐
void SpawnVoxels(MQObject lpVoxelized,const std::vector<VOXEL>& voxels,float fSize,MQPoint &vMin)
{
	for(unsigned int i=0;i<voxels.size();i++){
		float x=voxels[i].x * fSize + vMin.x;
		float y=voxels[i].y * fSize + vMin.y;
		float z=voxels[i].z * fSize + vMin.z;

		float s=fSize;

		int vertices[]={
			lpVoxelized->AddVertex(MQPoint(x  ,y  ,z  )),	//0
			lpVoxelized->AddVertex(MQPoint(x+s,y  ,z  )),	//1
			lpVoxelized->AddVertex(MQPoint(x+s,y+s,z  )),	//2
			lpVoxelized->AddVertex(MQPoint(x  ,y+s,z  )),	//3
			lpVoxelized->AddVertex(MQPoint(x  ,y  ,z+s)),	//4
			lpVoxelized->AddVertex(MQPoint(x+s,y  ,z+s)),	//5
			lpVoxelized->AddVertex(MQPoint(x+s,y+s,z+s)),	//6
			lpVoxelized->AddVertex(MQPoint(x  ,y+s,z+s)),	//7
		};

		int indices[]={
			4,5,1,0,//0,1,5,4,
			1,2,3,0,//0,3,2,1,
			3,7,4,0,//0,4,7,3,
			7,6,5,4,//4,5,6,7,
			6,7,3,2,//2,3,7,6,
			5,6,2,1,//1,2,6,5,
		};

		for(int j=0;j<COUNTOF(indices);j++){
			indices[j]=vertices[indices[j]];
		}

		for(int j=0;j<6;j++){
			// ʂǉ
			lpVoxelized->AddFace(4, &indices[j*4]);
		}
	}
}

//_̂ݐ;pr
void SpawnVertices(MQObject lpVoxelized,const std::vector<VOXEL>& voxels,float fSize,MQPoint &vMin)
{
	for(unsigned int i=0;i<voxels.size();i++){
		float x=voxels[i].x * fSize + vMin.x + fSize*0.5f;
		float y=voxels[i].y * fSize + vMin.y + fSize*0.5f;
		float z=voxels[i].z * fSize + vMin.z + fSize*0.5f;

		int idx=lpVoxelized->AddVertex(MQPoint(x  ,y  ,z  ));

		lpVoxelized->AddFace(1,&idx);
	}
}

